/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.framework.server;

import java.util.*;


import java.util.logging.Logger;
import java.util.logging.Level;

import org.apache.commons.lang.builder.ToStringBuilder;

import com.funambol.framework.logging.Sync4jLogger;
import com.funambol.framework.security.Sync4jPrincipal;



/**
 * Represent the mappings of a principal for a syncsource.
 *
 * @version $Id: ClientMapping.java,v 1.2 2006/04/09 20:19:20 nichele Exp $
 */
public class ClientMapping {

    //Contain a list of ClientMappingEntry
    private List clientMapping = new ArrayList();

    //Contain the principal of the client device
    private Sync4jPrincipal principal = null;

    // the requested database
    private String dbURI;

    //When the Client mapping are modified or added
    private boolean modified     = false;
    private List    modifiedKeys = new ArrayList();

    //When entries in the mapping are deleted
    private boolean deleted     = false;
    private List    deletedKeys = new ArrayList();

    // Keep a list of the entry to delete when the status is received
    private List    mappingToDelete = new ArrayList();

    //Transient keyword will disable serialisation for this member
    private transient Logger log = Sync4jLogger.getLogger();

    // ------------------------------------------------------------- Contructors

    /**
     * Construct a client mapping to a device Id the mapping must be populate
     * with data from the persistence store by calling
     * PersistenceStoreManager.read.
     *
     */
    public ClientMapping(Sync4jPrincipal principal, String dbURI) {
        this.principal = principal;
        this.dbURI     = dbURI    ;
    }

    // ---------------------------------------------------------- Public methods

    /**
     * @return true if the mapping where modified or added since initialization
     */
    public boolean isModified() {
        return modified;
    }

    /**
     * @return true if item where deleted since the initialisation
     */
    public boolean isDeleted() {
        return deleted;
    }

    /**
     * Initialize the client mapping with data from the Persistence Store
     * @param mapping A structure that contain GUID -> LUID
     */
    public void initializeFromMapping(List mapping) {
        resetMapping();
        clientMapping.addAll(mapping);
    }

    /**
     * Get the deleted LUIDs from the client mapping
     * @return a String[] of LUIDs deleted
     */
    public String[] getDeletedLuids() {
        if (deleted) {
            return (String[])deletedKeys.toArray(new String[deletedKeys.size()]);
        }

        return new String[0];
    }

    /**
     * Get the modified LUIDs from the client mapping
     * @return a String[] of LUIDs modified
     */
    public String[] getModifiedLuids() {
        if (modified) {
            return (String[])modifiedKeys.toArray(new String[modifiedKeys.size()]);
        }

        return new String[0];
    }

    /**
     * Returns the lastAnchor of the mapping identified with the given GUID
     * @param guid String
     * @return String
     */
    public String getLastAnchor(String guid) {
        ClientMappingEntry mappingEntry = getClientMappingEntry(guid);
        if (mappingEntry == null) {
            return null;
        }
        return mappingEntry.getLastAnchor();
    }


    /**
     * Get the current mapping.
     *
     * @return the current mapping as a List
     */
    public List getMapping() {
        List ret = new ArrayList();
        ret.addAll(clientMapping);
        return ret;
    }

    /**
     * Returns a Map with the valid GUID-LUID mapping.
     *
     * @return a Map with the valid GUID-LUID mapping.
     */
    public Map getValidMapping() {

        Map                result         = new HashMap();
        Iterator           itMappingEntry = clientMapping.iterator();
        ClientMappingEntry mappingEntry   = null;

        while (itMappingEntry.hasNext()) {
            mappingEntry = (ClientMappingEntry)itMappingEntry.next();
            if (mappingEntry.isValid()) {
                result.put(mappingEntry.getGuid(), mappingEntry.getLuid());
            }
        }
        return result;
    }

    /**
     * Get the Client Device Id that correspond to that Client
     * @return The Device Id
     */
    public String getClientDeviceId() {
        return principal.getDeviceId();
    }

    /**
     * Get the principal
     *
     * @return the principal
     */
    public Sync4jPrincipal getPrincipal() {
        return principal;
    }

    /**
     * Get the database uri
     *
     * @return <i>dbURI</i>
     */
    public String getDbURI() {
        return dbURI;
    }

    /**
     * Add a mapping between a GUID and a LUID from the server
     * @param guid The GUID of the server item
     * @param luid The LUID of the client item
     */
    private void addMapping(String guid, String luid, String lastAnchor) {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Adding mapping LUID-GUID '" + luid + "-" + guid + "' " +
                       "with lastAnchor '" + lastAnchor + "'");
        }

        //
        // if the same guid was already mapped to a luid, we have to remeber
        // the former one and remove it from the list of updated/deleted luids
        //
        ClientMappingEntry newEntry = new ClientMappingEntry(guid, luid, lastAnchor);
        modifiedKeys.add(luid); modified = true;

        int index = clientMapping.indexOf(newEntry);
        if (index != -1) {
            ClientMappingEntry oldMapping =
                (ClientMappingEntry)clientMapping.set(index, newEntry);

            String oldLuid = oldMapping.getLuid();
            if (oldLuid != null) {
                  removeDeletedKey(oldLuid);
                  removeModifiedKey(oldLuid);
              }
        } else {
            clientMapping.add(newEntry);
        }
    }

    /**
     * Get the mapped GUID for the given LUID.
     * @param luid The LUID of the client item
     * @return The mapped value for the key in input
     */
    public String getMappedValueForLuid(String luid) {
        String result = null;

        Iterator           itMappingEntry = clientMapping.iterator();
        ClientMappingEntry mappingEntry   = null;

        while (itMappingEntry.hasNext()) {
            mappingEntry = (ClientMappingEntry)itMappingEntry.next();
            if (mappingEntry.getLuid().equals(luid)) {
                return mappingEntry.getGuid();
            }
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("No mapping found for LUID: " + luid);
        }
        return result;
    }

    /**
     * Get the mapped LUID key for the given GUID value.
     * @param guid The GUID of the client item
     * @return The mapped value for the key in input
     */
    public String getMappedValueForGuid(String guid) {
        String luid = null;
        ClientMappingEntry mappingEntry =
            new ClientMappingEntry(guid, null, null);

        int index = clientMapping.indexOf(mappingEntry);
        if (index == -1) {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("No mapping found for GUID: " + guid);
            }
        } else {
            mappingEntry = (ClientMappingEntry)clientMapping.get(index);
            luid = mappingEntry.getLuid();
        }

        return luid;
    }

    /**
     * Remove a mapped values from the cache given a GUID
     * @param guid The GUID for the item from the client
     * @param isDiffered if true the mapping is just removed from the clientMapping,
     *                   if false the mapping is removed from the database
     */
    public void removeMappedValuesForGuid(String guid, boolean isDiffered) {

        if (isDiffered) {
            ClientMappingEntry entry = getClientMappingEntry(guid);

            if (entry != null) {
                if (!mappingToDelete.contains(entry)) {
                    mappingToDelete.add(entry);
                }
                clientMapping.remove(guid);
            }
        } else {

            ClientMappingEntry entry = new ClientMappingEntry(guid);

            int index = mappingToDelete.indexOf(entry);
            if (index != -1) {
                //
                // There is a previous entry deleted with isDiffered = true
                //
                entry = (ClientMappingEntry)mappingToDelete.get(index);
                mappingToDelete.remove(index);
                addDeletedKey(entry.getLuid());

            }
            //
            // BTW, we check is there is a mapping in the current clientMapping
            //
            entry = getClientMappingEntry(guid);
            if (entry != null) {
                clientMapping.remove(entry);
                addDeletedKey(entry.getLuid());
            }
        }
    }

    /**
     * Remove a mapped values given a GUID only if the mapping is already removed
     * with isDiffered = true.
     * @param guid The GUID for the item from the client
     */
    public void confirmRemoveMappedValuesForGuid(String guid) {

        ClientMappingEntry entry = new ClientMappingEntry(guid);

        int index = mappingToDelete.indexOf(entry);
        if (index != -1) {
            //
            // There is a previous entry deleted with isDiffered = true
            //
            entry = (ClientMappingEntry)mappingToDelete.get(index);
            addDeletedKey(entry.getLuid());

        }
    }

    /**
     * Updates a key value mapping. If the mapping does not exist, it calls
     * <code>addMapping(guid, luid)</code>, otherwise the existing mapping is
     * updated.
     *
     * @param guid The GUID value from the server item
     * @param luid The LUID value from the client item
     * @param lastAnchor The lastAnchor of the mapping
     */
    public void updateMapping(String guid, String luid, String lastAnchor) {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Updating mapping LUID-GUID '" + luid + "-" + guid + "' " +
                       "with lastAnchor '" + lastAnchor + "'");
        }

        ClientMappingEntry mappingEntry = new ClientMappingEntry(guid, luid, lastAnchor);

        //
        // If new, just call addMapping
        //
        int index = clientMapping.indexOf(mappingEntry);
        if (index == -1) {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("The mapping is new.");
            }
            addMapping(guid, luid, lastAnchor);
            return;
        }

        //
        // if the same guid was already mapped to a luid, we have to remember
        // the former one and remove it from the list of updated/deleted luids
        //
        ClientMappingEntry oldMapping =
            (ClientMappingEntry)clientMapping.set(index, mappingEntry);
        String oldLuid = oldMapping.getLuid();
        modifiedKeys.add(luid); modified = true;

        if (oldLuid != null) {
            removeDeletedKey(oldLuid);
            removeModifiedKey(oldLuid);
        }
        //
        // If we add an entry that was considered deleted,
        // remove it from the deleted entries.
        //
        removeDeletedKey(luid);
    }


    /**
     * Updates a key value mapping. If the mapping does not exist, it calls
     * <code>addMapping(guid, luid)</code>, otherwise the existing mapping is
     * updated.
     *
     * @param guid The GUID value from the server item
     * @param luid The LUID value from the client item
     */
    public void updateLastAnchor(String guid, String luid, String lastAnchor) {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Updating mapping LUID-GUID '" + luid + "-" + guid + "' " +
                       "with lastAnchor '" + lastAnchor + "'");
        }

        ClientMappingEntry mappingEntry = new ClientMappingEntry(guid, luid, null);

        int index = clientMapping.indexOf(mappingEntry);
        if (index == -1) {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Mapping not found.");
            }
            return;
        }
        mappingEntry = (ClientMappingEntry)clientMapping.get(index);

        if (luid.equals(mappingEntry.getLuid())) {
            mappingEntry.setLastAnchor(lastAnchor);
            modifiedKeys.add(mappingEntry.getLuid()); modified = true;
        } else {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("There is a mapping with GUID '" + guid + "' but with " +
                    " different luid (" + mappingEntry.getLuid() + ")");
            }
        }
    }

    /**
     * Checks if the mapping with the given guid is valid.
     * @param guid the guid of the mapping to check
     * @return true if the mapping is valid, false if the mapping isn't valid or
     *         it doesn't exist.
     */
    public boolean isValid(String guid) {
        ClientMappingEntry mappingEntry = getClientMappingEntry(guid);
        if (mappingEntry == null) {
            return false;
        }
        return mappingEntry.isValid();
    }

    /**
     * Clears these mappings moving the existing and new mappings to deleted.
     * This is used for example to re-initialize the LUID-GUID mapping for a
     * slow sync.
     */
    public void clearMappings() {
        Iterator it = clientMapping.iterator();
        while (it.hasNext()) {
            deletedKeys.add(((ClientMappingEntry)it.next()).getLuid());
        }

        deletedKeys.addAll(modifiedKeys);

        clientMapping.clear();

        modified = false;
        deleted = true;
    }

    public String toString() {
         ToStringBuilder sb = new ToStringBuilder(this);

         sb.append("clientMapping", clientMapping);
         sb.append("modifiedKeys" , modifiedKeys );
         sb.append("deletedKeys"  , deletedKeys  );

         return sb.toString();
    }

    /**
     * Resets the modified (updated/deleted) uids
     */
    public void resetModifiedKeys() {
        //If we have modified item to the map
        if (modified) {
            modifiedKeys.clear();
            modified = false;
        }
        //If we have deleted item to the map
        if (deleted) {
            deletedKeys.clear();
            deleted = false;
        }
    }

    /**
     * Get the current ClientMappingEntry for the given guid.
     *
     * @return the current mapping ClientMappingEntry
     */
    public ClientMappingEntry getClientMappingEntry(String guid) {
        ClientMappingEntry entry = new ClientMappingEntry(guid);
        int index = clientMapping.indexOf(entry);
        if (index == -1) {
            return null;
        }
        entry = (ClientMappingEntry)clientMapping.get(index);
        return entry;

    }

    /**
     * Get the current ClientMappingEntry for the given luid.
     *
     * @return the current mapping ClientMappingEntry
     */
    public ClientMappingEntry getClientMappingEntryByLuid(String luid) {

        Iterator           itMappingEntry = clientMapping.iterator();
        ClientMappingEntry mappingEntry   = null;

        while (itMappingEntry.hasNext()) {
            mappingEntry = (ClientMappingEntry)itMappingEntry.next();
            if (mappingEntry.getLuid().equals(luid)) {
                return mappingEntry;
            }
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("No mapping found for LUID: " + luid);
        }
        return null;

    }

    // --------------------------------------------------------- Private methods

    /**
     * Reset the content of the Client mapping list and the state
     * of the current mapping.
     */
    private void resetMapping() {
        //We clean the list
        if (!clientMapping.isEmpty()) {
            clientMapping.clear();
        }

        resetModifiedKeys();
    }

    /**
     * Remove LUID from the deleted entries
     * @param luid The LUID for the item from the client
     */
    private void removeDeletedKey(String luid) {
        if (deleted && deletedKeys.contains(luid)) {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Removing deleted LUID '" + luid + "' from deletedKeys");
            }
            deletedKeys.remove(luid);
            if (deletedKeys.size() == 0) {
                deleted = false;
            }
        }
    }

    /**
     * Remove LUID from the modified entries
     * @param luid The LUID for the item from the client
     */
    private void removeModifiedKey(String luid) {
        if (modified && modifiedKeys.contains(luid)) {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Removing modified LUID '" + luid + "' from modifiedKeys");
            }
            modifiedKeys.remove(luid);
            if (modifiedKeys.size() == 0) {
                modified = false;
            }
        }
    }

    /**
     * Add LUID into the deleted keys.
     * @param luid The LUID for the item from the client
     */
    private void addDeletedKey(String luid) {
        removeModifiedKey(luid);
        deleted = true;
        deletedKeys.add(luid);
    }

}
