/**
 * Copyright (C) 2005-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.funambol.framework.server.store;

import java.io.Serializable;
import java.util.ArrayList;

import com.funambol.framework.filter.*;

/**
 * This class contains the methods to convert a Clause into a PreparedWhere.
 *
 * @version $Id: SQLHelperClause.java,v 1.1 2006/01/18 17:00:11 luigiafassina Exp $
 *
 */
public class SQLHelperClause  {

    // --------------------------------------------------------------- Constants
    private static final String OPT_UPPER      = "upper("    ;
    private static final String OPT_CLOSE     = ")"                 ;
    private static final String OPT_UPPER_DB2 = "UPPER(CAST("       ;
    private static final String OPT_CLOSE_DB2 = " AS VARCHAR(255)))";
    private static final String PRODUCT_DB2   = "DB2"               ;

    // ------------------------------------------------------------ Private data
    private String uprOpen     = null;
    private String uprClose    = null;
    private String uprProperty = null;
    // ------------------------------------------------------------ Constructors
    public SQLHelperClause(String dbProductName) {

        if (dbProductName != null &&
            dbProductName.toUpperCase().indexOf(PRODUCT_DB2) != -1) {

            uprOpen  = OPT_UPPER_DB2;
            uprClose = OPT_CLOSE_DB2;
        } else {
            uprOpen  = OPT_UPPER;
            uprClose = OPT_CLOSE;
        }
    }

    // ---------------------------------------------------------- Public Methods

    /**
     * Creates a PreparedWhere from the given Clause
     *
     * @param clause the clause
     * @return a PreparedWhered related to the given Clause
     */
    public PreparedWhere getPreparedWhere(Clause clause) {

        ClassLoader loader = Thread.currentThread().getContextClassLoader();
        ClassLoader clauseLoader = clause.getClass().getClassLoader();

        if (clause instanceof com.funambol.framework.filter.LogicalClause) {
            return getPreparedWhere( (LogicalClause)clause);
        } else if (clause instanceof com.funambol.framework.filter.LogicalClause) {
            return getPreparedWhere( (LogicalClause)clause);
        } else if (clause instanceof com.funambol.framework.filter.WhereClause) {
            return getPreparedWhere( (WhereClause)clause);
        } else if (clause instanceof com.funambol.framework.filter.AllClause) {
            return getPreparedWhere( (AllClause)clause);
        } else if (clause instanceof com.funambol.framework.filter.FieldClause) {
            //
            // The FieldClause aren't converted in PreparedWhere
            //
            return getEmptyPreparedWhere();
        }

        if (clause == null) {
            throw new IllegalStateException("Clause is null");
        }

        throw new IllegalStateException("Unknown clause: '" + clause.getClass().getName() + "'");
    }

    /**
     * Creates a PreparedWhere from the given LogicalClause
     *
     * @param clause the logical clause
     * @return a PreparedWhered related to the given LogicalClause
     */
    public PreparedWhere getPreparedWhere(LogicalClause clause) {

        StringBuffer sql = new StringBuffer();
        ArrayList parameters = new ArrayList();
        PreparedWhere pw = null;

        if (clause.isUnaryOperator()) {
            pw = getPreparedWhere(clause.getOperands()[0]);
            sql.append(clause.getOperator().toLowerCase()).append(pw.sql);
            for(int i=0; i<pw.parameters.length; ++i) {
                parameters.add(pw.parameters[i]);
            }
        } else {
            Clause[] operands = clause.getOperands();
            for (int i=0; i<operands.length; ++i) {
                pw = getPreparedWhere(operands[i]);

                if (i>0) {
                    sql.append(' ').append(clause.getOperator().toLowerCase()).append(' ');
                }
                sql.append(pw.sql);

                for(int j=0; j<pw.parameters.length; ++j) {
                    parameters.add(pw.parameters[j]);
                }
            }
        }

        PreparedWhere ret = new PreparedWhere();

        ret.sql = '(' + sql.toString() + ')';
        ret.parameters = parameters.toArray();

        return ret;

    }

    /**
     * Creates a PreparedWhere from the given WhereClause
     *
     * @param clause the where clause
     * @return a PreparedWhered related to the given WhereClause
     */
    public PreparedWhere getPreparedWhere(WhereClause clause) {
        String property = clause.getProperty();
        String operator = clause.getOperator();
        String[] values = clause.getValues();
        boolean caseSensitive = clause.isCaseSensitive();

        StringBuffer query = new StringBuffer();

        assert (values != null);

        PreparedWhere where = new PreparedWhere();
        where.parameters = new Object[values.length];

        if (!caseSensitive) {
            uprProperty = " UPPER("+property+")";
        } else {
            uprOpen     = ""      ;
            uprClose    = ""      ;
            uprProperty = property;
        }

        if (WhereClause.OPT_START_WITH.equalsIgnoreCase(operator)) {
            query.append(uprProperty).append(" LIKE ").append(uprOpen).append('?').append(uprClose);
            where.sql = query.toString();
            where.parameters[0] = values[0] + '%';
        } else if (WhereClause.OPT_END_WITH.equalsIgnoreCase(operator)) {
            query.append(uprProperty).append(" LIKE ").append(uprOpen).append('?').append(uprClose);
            where.sql = query.toString();
            where.parameters[0] = '%' + values[0];
        } else if (WhereClause.OPT_CONTAINS.equalsIgnoreCase(operator)) {
            query.append(uprProperty).append(" LIKE ").append(uprOpen).append('?').append(uprClose);
            where.sql = query.toString();
            where.parameters[0] = '%' + values[0] + '%';
        } else if (WhereClause.OPT_EQ.equalsIgnoreCase(operator)) {
            query.append(uprProperty).append(" = ").append(uprOpen).append('?').append(uprClose);
            where.sql = query.toString();
            where.parameters[0] = values[0];
        } else if (WhereClause.OPT_GT.equalsIgnoreCase(operator)) {
            query.append(uprProperty).append(" > ").append(uprOpen).append('?').append(uprClose);
            where.sql = query.toString();
            where.parameters[0] = values[0];
        } else if (WhereClause.OPT_LT.equalsIgnoreCase(operator)) {
            query.append(uprProperty).append(" < ").append(uprOpen).append('?').append(uprClose);
            where.sql = query.toString();
            where.parameters[0] = values[0];
        } else if (WhereClause.OPT_BETWEEN.equalsIgnoreCase(operator)) {
            query.append(uprProperty).append(' ')
                .append(WhereClause.OPT_BETWEEN).append(' ').append(uprOpen)
                .append('?').append(uprClose)
                .append(" AND ").append(uprOpen).append('?').append(uprClose);
            where.sql = query.toString();
            where.parameters[0] = values[0];
            where.parameters[1] = values[1];
        } else if (WhereClause.OPT_GE.equalsIgnoreCase(operator)) {
            query.append(uprProperty).append(" >= ").append(uprOpen).append('?').append(uprClose);
            where.sql = query.toString();
            where.parameters[0] = values[0];
        } else if (WhereClause.OPT_LE.equalsIgnoreCase(operator)) {
            query.append(uprProperty).append(" <= ").append(uprOpen).append('?').append(uprClose);
            where.sql = query.toString();
            where.parameters[0] = values[0];
        }

        where.sql = '(' + where.sql + ')';
        return where;
    }


    /**
     * Creates a PreparedWhere from the given AllClause
     *
     * @param clause the where clause
     * @return a PreparedWhered related to the given AllClause
     */
    public PreparedWhere getPreparedWhere(AllClause clause) {
        PreparedWhere ret = new PreparedWhere();

        ret.sql = "";
        ret.parameters = new Object[0];

        return ret;
    }

    /**
     * Creates a empty PreparedWhere
     *
     * @return a empty PreparedWhere
     */
    public PreparedWhere getEmptyPreparedWhere() {
        PreparedWhere ret = new PreparedWhere();

        ret.sql = "";
        ret.parameters = new Object[0];

        return ret;
    }

}
