/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.funambol.transport.http.server;

import java.util.Map;
import java.util.HashMap;

import javax.servlet.http.*;
import javax.naming.*;
import javax.ejb.CreateException;
import javax.ejb.RemoveException;
import javax.ejb.EJBException;
import javax.rmi.PortableRemoteObject;

import com.funambol.framework.transport.http.SyncHolder;
import com.funambol.framework.server.SyncResponse;
import com.funambol.server.syncbean.SyncLocal;
import com.funambol.server.syncbean.SyncHomeLocal;

import com.funambol.framework.core.Sync4jException;
import com.funambol.framework.server.error.ServerException;
import com.funambol.framework.server.error.ServerFailureException;
import com.funambol.framework.server.error.NotImplementedException;
import com.funambol.framework.protocol.ProtocolException;


import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;

/**
 * Implementes a <i>SyncHolder</i> wrapping a local EJB.
 *
 *
 * @version $Id: LocalEJBSyncHolder.java,v 1.2 2006/04/29 15:36:53 nichele Exp $
 */
public class LocalEJBSyncHolder implements SyncHolder {

    // --------------------------------------------------------------- Constants

    public static final String SYNCEJB_HOME_JNDI_NAME =
        "java:comp/env/ejb/LocalSyncBean";

    // ------------------------------------------------------------ Private data

    private SyncLocal     sync              = null;
    private String        sessionId         = null;
    private SyncHomeLocal home              = null;
    private long          creationTimestamp       ;

    // ------------------------------------------------------------ Constructors

    public LocalEJBSyncHolder() throws NamingException {
        InitialContext ctx = new InitialContext();
        home = (SyncHomeLocal)PortableRemoteObject.narrow(
            ctx.lookup(SYNCEJB_HOME_JNDI_NAME),
            SyncHomeLocal.class
        );
        creationTimestamp = System.currentTimeMillis();
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Processes an incoming XML message.
     *
     * @param requestURI the uri of the request
     * @param requestData the SyncML request as an array of bytes
     * @param parameters SyncML request parameters
     * @param headers SyncML request headers
     *
     * @return the SyncML response as a <i>ISyncResponse</i> object
     *
     * @throws ServerException in case of a server error
     *
     */
    public SyncResponse processXMLMessage(final String requestURI,
                                          final byte[] requestData,
                                          final Map    parameters ,
                                          final Map    headers    )
    throws ServerException {
        return sync.processXMLMessage(requestURI, requestData, parameters, headers);
    }

    /**
     * Processes an incoming WBXML message.
     *
     * @param requestURI the uri of the request
     * @param requestData the SyncML request as an array of bytes
     * @param parameters SyncML request parameters
     * @param headers SyncML request headers
     *
     * @return the SyncML response as a <i>ISyncResponse</i> object
     *
     * @throws ServerException in case of a server error
     *
     */
    public SyncResponse processWBXMLMessage(final String requestURI,
                                            final byte[] requestData,
                                            final Map    parameters ,
                                            final Map    headers    )
    throws ServerException {
        return sync.processWBXMLMessage(requestURI, requestData, parameters, headers);
    }

    public void setSessionId(String sessionId) throws Sync4jException {
        try {
            sync = home.create(sessionId);
        } catch (CreateException e) {
            throw new ServerFailureException(e);
        }

        this.sessionId = sessionId;
    }

    public String getSessionId() {
        return sessionId;
    }

    /** Called when the SyncHolder is not required any more. It gives the holder
     * an opportunity to release and clean up resources.
     *
     * @throws java.lang.Exception in case of error. The real exception is stored
     * in the cause.
     *
     */
    public void close() throws Exception {
        try {
            sync.remove();
        } catch (RemoveException e) {
            throw new Exception("Error in closing the SyncHolder", e);
        } catch (EJBException e) {
            throw new Exception("Error in closing the SyncHolder", e);
        }
    }

    /**
     * Returns the creation timestamp (in milliseconds since midnight, January
     * 1, 1970 UTC).
     *
     * @see com.funambol.framework.transport.http.SyncHolder
     */
    public long getCreationTimestamp() {
        return creationTimestamp;
    }
}
